from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _
from django.http import HttpRequest, JsonResponse, Http404
from django.shortcuts import render

from intervention.forms.forms import NewInterventionForm, EditInterventionForm
from intervention.forms.modalForms import ShareModalForm, NewRevocationModalForm, \
    CheckModalForm, NewDeductionModalForm, NewInterventionDocumentModalForm, RemoveEcoAccountDeductionModalForm, \
    RemoveRevocationModalForm, EditEcoAccountDeductionModalForm, EditRevocationModalForm
from intervention.models import Intervention, Revocation, InterventionDocument, RevocationDocument
from intervention.tables import InterventionTable
from konova.contexts import BaseContext
from konova.decorators import *
from konova.forms import SimpleGeomForm, RemoveModalForm, RecordModalForm, EditDocumentModalForm
from konova.sub_settings.context_settings import TAB_TITLE_IDENTIFIER
from konova.utils.documents import remove_document, get_document
from konova.utils.generators import generate_qr_code
from konova.utils.message_templates import INTERVENTION_INVALID, FORM_INVALID, IDENTIFIER_REPLACED, \
    CHECKED_RECORDED_RESET, DEDUCTION_REMOVED, DEDUCTION_ADDED, REVOCATION_ADDED, REVOCATION_REMOVED, \
    COMPENSATION_REMOVED_TEMPLATE, DOCUMENT_ADDED, DEDUCTION_EDITED, REVOCATION_EDITED, DOCUMENT_EDITED, \
    RECORDED_BLOCKS_EDIT
from konova.utils.user_checks import in_group


@login_required
@any_group_check
def index_view(request: HttpRequest):
    """
    Renders the index view for Interventions

    Args:
        request (HttpRequest): The incoming request

    Returns:
        A rendered view
    """
    template = "generic_index.html"

    # Filtering by user access is performed in table filter inside of InterventionTableFilter class
    interventions = Intervention.objects.filter(
        deleted=None,  # not deleted
    ).select_related(
        "legal"
    )
    table = InterventionTable(
        request=request,
        queryset=interventions
    )
    context = {
        "table": table,
        TAB_TITLE_IDENTIFIER: _("Interventions - Overview"),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
def new_view(request: HttpRequest):
    """
    Renders a view for a new intervention creation

    Args:
        request (HttpRequest): The incoming request

    Returns:

    """
    template = "intervention/form/view.html"
    data_form = NewInterventionForm(request.POST or None)
    geom_form = SimpleGeomForm(request.POST or None, read_only=False)
    if request.method == "POST":
        if data_form.is_valid() and geom_form.is_valid():
            generated_identifier = data_form.cleaned_data.get("identifier", None)
            intervention = data_form.save(request.user, geom_form)
            if generated_identifier != intervention.identifier:
                messages.info(
                    request,
                    IDENTIFIER_REPLACED.format(
                        generated_identifier,
                        intervention.identifier
                    )
                )
            messages.success(request, _("Intervention {} added").format(intervention.identifier))
            return redirect("intervention:detail", id=intervention.id)
        else:
            messages.error(request, FORM_INVALID, extra_tags="danger",)
    else:
        # For clarification: nothing in this case
        pass
    context = {
        "form": data_form,
        "geom_form": geom_form,
        TAB_TITLE_IDENTIFIER: _("New intervention"),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
def new_id_view(request: HttpRequest):
    """ JSON endpoint

    Provides fetching of free identifiers for e.g. AJAX calls

    """
    tmp_intervention = Intervention()
    identifier = tmp_intervention.generate_new_identifier()
    while Intervention.objects.filter(identifier=identifier).exists():
        identifier = tmp_intervention.generate_new_identifier()
    return JsonResponse(
        data={
            "gen_data": identifier
        }
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def new_document_view(request: HttpRequest, id: str):
    """ Renders a form for uploading new documents

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id to which the new document will be related
    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    form = NewInterventionDocumentModalForm(request.POST or None, request.FILES or None, instance=intervention, request=request)
    return form.process_request(
        request,
        msg_success=DOCUMENT_ADDED,
        redirect_url=reverse("intervention:detail", args=(id,)) + "#related_data"
    )


@login_required
@default_group_required
def get_revocation_view(request: HttpRequest, doc_id: str):
    """ Returns the revocation document as downloadable file

    Wraps the generic document fetcher function from konova.utils.

    Args:
        request (HttpRequest): The incoming request
        doc_id (str): The document id

    Returns:

    """
    doc = get_object_or_404(RevocationDocument, id=doc_id)
    # File download only possible if related instance is shared with user
    if not doc.instance.legal.intervention.users.filter(id=request.user.id):
        messages.info(
            request,
            DATA_UNSHARED
        )
        return redirect("intervention:detail", id=doc.instance.id)
    return get_document(doc)

@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def get_document_view(request: HttpRequest, id: str, doc_id: str):
    """ Returns the document as downloadable file

    Wraps the generic document fetcher function from konova.utils.

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention id
        doc_id (str): The document id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    doc = get_object_or_404(InterventionDocument, id=doc_id)
    return get_document(doc)


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def remove_document_view(request: HttpRequest, id: str, doc_id: str):
    """ Removes the document from the database and file system

    Wraps the generic functionality from konova.utils.

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention id
        doc_id (str): The document id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    doc = get_object_or_404(InterventionDocument, id=doc_id)
    return remove_document(
        request,
        doc
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def edit_document_view(request: HttpRequest, id: str, doc_id: str):
    """ Removes the document from the database and file system

    Wraps the generic functionality from konova.utils.

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention id
        doc_id (str): The document id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    doc = get_object_or_404(InterventionDocument, id=doc_id)
    form = EditDocumentModalForm(request.POST or None, request.FILES or None, instance=intervention, document=doc, request=request)
    return form.process_request(
        request,
        DOCUMENT_EDITED,
        redirect_url=reverse("intervention:detail", args=(intervention.id,)) + "#related_data"
    )


@login_required
@any_group_check
def detail_view(request: HttpRequest, id: str):
    """ Renders a detail view for viewing an intervention's data

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id

    Returns:

    """
    template = "intervention/detail/view.html"

    # Fetch data, filter out deleted related data
    intervention = get_object_or_404(
        Intervention.objects.select_related(
            "geometry",
            "legal",
            "responsible",
        ),
        id=id
    )
    compensations = intervention.compensations.filter(
        deleted=None,
    )
    _user = request.user
    is_data_shared = intervention.is_shared_with(user=_user)

    geom_form = SimpleGeomForm(
        instance=intervention,
    )

    parcels = intervention.get_underlying_parcels()

    context = {
        "obj": intervention,
        "compensations": compensations,
        "has_access": is_data_shared,
        "geom_form": geom_form,
        "parcels": parcels,
        "is_default_member": in_group(_user, DEFAULT_GROUP),
        "is_zb_member": in_group(_user, ZB_GROUP),
        "is_ets_member": in_group(_user, ETS_GROUP),
        "LANIS_LINK": intervention.get_LANIS_link(),
        TAB_TITLE_IDENTIFIER: f"{intervention.identifier} - {intervention.title}",
    }

    request = intervention.set_status_messages(request)

    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def edit_view(request: HttpRequest, id: str):
    """
    Renders a view for editing interventions

    Args:
        request (HttpRequest): The incoming request

    Returns:

    """
    template = "intervention/form/view.html"
    # Get object from db
    intervention = get_object_or_404(Intervention, id=id)
    if intervention.is_recorded:
        messages.info(
            request,
            RECORDED_BLOCKS_EDIT
        )
        return redirect("intervention:detail", id=id)

    # Create forms, initialize with values from db/from POST request
    data_form = EditInterventionForm(request.POST or None, instance=intervention)
    geom_form = SimpleGeomForm(request.POST or None, read_only=False, instance=intervention)
    if request.method == "POST":
        if data_form.is_valid() and geom_form.is_valid():
            # The data form takes the geom form for processing, as well as the performing user
            # Save the current state of recorded|checked to inform the user in case of a status reset due to editing
            i_rec = intervention.recorded is not None
            i_check = intervention.checked is not None
            intervention = data_form.save(request.user, geom_form)
            messages.success(request, _("Intervention {} edited").format(intervention.identifier))
            if i_check or i_rec:
                messages.info(request, CHECKED_RECORDED_RESET)
            return redirect("intervention:detail", id=intervention.id)
        else:
            messages.error(request, FORM_INVALID, extra_tags="danger",)
    else:
        # For clarification: nothing in this case
        pass
    context = {
        "form": data_form,
        "geom_form": geom_form,
        TAB_TITLE_IDENTIFIER: _("Edit {}").format(intervention.identifier),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def remove_view(request: HttpRequest, id: str):
    """ Renders a remove view for this intervention

    Args:
        request (HttpRequest): The incoming request
        id (str): The uuid id as string

    Returns:

    """
    obj = Intervention.objects.get(id=id)
    identifier = obj.identifier
    form = RemoveModalForm(request.POST or None, instance=obj, request=request)
    return form.process_request(
        request,
        _("{} removed").format(identifier),
        redirect_url=reverse("intervention:index")
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def edit_revocation_view(request: HttpRequest, id: str, revocation_id: str):
    """ Renders a edit view for a revocation

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id as string
        revocation_id (str): The revocation's id as string

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    revocation = get_object_or_404(Revocation, id=revocation_id)

    form = EditRevocationModalForm(request.POST or None, request.FILES or None, instance=intervention, revocation=revocation, request=request)
    return form.process_request(
        request,
        REVOCATION_EDITED,
        redirect_url=reverse("intervention:detail", args=(intervention.id,)) + "#related_data"
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def remove_revocation_view(request: HttpRequest, id: str, revocation_id: str):
    """ Renders a remove view for a revocation

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id as string
        revocation_id (str): The revocation's id as string

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    revocation = get_object_or_404(Revocation, id=revocation_id)

    form = RemoveRevocationModalForm(request.POST or None, instance=intervention, revocation=revocation, request=request)
    return form.process_request(
        request,
        REVOCATION_REMOVED,
        redirect_url=reverse("intervention:detail", args=(intervention.id,)) + "#related_data"
    )


@login_required
def share_view(request: HttpRequest, id: str, token: str):
    """ Performs sharing of an intervention

    If token given in url is not valid, the user will be redirected to the dashboard

    Args:
        request (HttpRequest): The incoming request
        id (str): Intervention's id
        token (str): Access token for intervention

    Returns:

    """
    user = request.user
    intervention = get_object_or_404(Intervention, id=id)
    # Check tokens
    if intervention.access_token == token:
        # Send different messages in case user has already been added to list of sharing users
        if intervention.is_shared_with(user):
            messages.info(
                request,
                _("{} has already been shared with you").format(intervention.identifier)
            )
        else:
            messages.success(
                request,
                _("{} has been shared with you").format(intervention.identifier)
            )
            intervention.share_with_user(user)
        return redirect("intervention:detail", id=id)
    else:
        messages.error(
            request,
            _("Share link invalid"),
            extra_tags="danger",
        )
        return redirect("home")


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def create_share_view(request: HttpRequest, id: str):
    """ Renders sharing form for an intervention

    Args:
        request (HttpRequest): The incoming request
        id (str): Intervention's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    form = ShareModalForm(request.POST or None, instance=intervention, request=request)
    return form.process_request(
        request,
        msg_success=_("Share settings updated")
    )


@login_required
@registration_office_group_required
@shared_access_required(Intervention, "id")
def check_view(request: HttpRequest, id: str):
    """ Renders check form for an intervention

    Args:
        request (HttpRequest): The incoming request
        id (str): Intervention's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    form = CheckModalForm(request.POST or None, instance=intervention, request=request)
    return form.process_request(
        request,
        msg_success=_("Check performed"),
        msg_error=INTERVENTION_INVALID
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def new_revocation_view(request: HttpRequest, id: str):
    """ Renders sharing form for an intervention

    Args:
        request (HttpRequest): The incoming request
        id (str): Intervention's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    form = NewRevocationModalForm(request.POST or None, request.FILES or None, instance=intervention, request=request)
    return form.process_request(
        request,
        msg_success=REVOCATION_ADDED,
        redirect_url=reverse("intervention:detail", args=(id,)) + "#related_data"
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def log_view(request: HttpRequest, id: str):
    """ Renders a log view using modal

    Args:
        request (HttpRequest): The incoming request
        id (str): The compensation's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    template = "modal/modal_generic.html"
    body_template = "log.html"

    context = {
        "modal_body_template": body_template,
        "log": intervention.log.all(),
        "modal_title": _("Log"),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def new_deduction_view(request: HttpRequest, id: str):
    """ Renders a modal form view for creating deductions

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id which shall benefit from this deduction

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    form = NewDeductionModalForm(request.POST or None, instance=intervention, request=request)
    return form.process_request(
        request,
        msg_success=DEDUCTION_ADDED,
        redirect_url=reverse("intervention:detail", args=(id,)) + "#related_data",
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def remove_deduction_view(request: HttpRequest, id: str, deduction_id: str):
    """ Renders a modal view for removing deductions

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id
        deduction_id (str): The deduction's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    try:
        eco_deduction = intervention.deductions.get(id=deduction_id)
    except ObjectDoesNotExist:
        raise Http404("Unknown deduction")

    form = RemoveEcoAccountDeductionModalForm(request.POST or None, instance=intervention, deduction=eco_deduction, request=request)
    return form.process_request(
        request=request,
        msg_success=DEDUCTION_REMOVED,
        redirect_url=reverse("intervention:detail", args=(id,)) + "#related_data"
    )


@login_required
@default_group_required
@shared_access_required(Intervention, "id")
def edit_deduction_view(request: HttpRequest, id: str, deduction_id: str):
    """ Renders a modal view for removing deductions

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id
        deduction_id (str): The deduction's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    try:
        eco_deduction = intervention.deductions.get(id=deduction_id)
    except ObjectDoesNotExist:
        raise Http404("Unknown deduction")

    form = EditEcoAccountDeductionModalForm(request.POST or None, instance=intervention, deduction=eco_deduction, request=request)
    return form.process_request(
        request=request,
        msg_success=DEDUCTION_EDITED,
        redirect_url=reverse("intervention:detail", args=(id,)) + "#related_data"
    )


@login_required
@conservation_office_group_required
@shared_access_required(Intervention, "id")
def record_view(request: HttpRequest, id: str):
    """ Renders a modal form for recording an intervention

    Args:
        request (HttpRequest): The incoming request
        id (str): The intervention's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    form = RecordModalForm(request.POST or None, instance=intervention, request=request)
    msg_succ = _("{} unrecorded") if intervention.recorded else _("{} recorded")
    msg_succ = msg_succ.format(intervention.identifier)
    return form.process_request(
        request,
        msg_succ,
        msg_error=_("There are errors on this intervention:")
    )


def remove_compensation_view(request:HttpRequest, id: str, comp_id: str):
    """ Renders a modal view for removing the compensation

    Args:
        request (HttpRequest): The incoming request
        id (str): The compensation's id

    Returns:

    """
    intervention = get_object_or_404(Intervention, id=id)
    try:
        comp = intervention.compensations.get(
            id=comp_id
        )
    except ObjectDoesNotExist:
        raise Http404("Unknown compensation")
    form = RemoveModalForm(request.POST or None, instance=comp, request=request)
    return form.process_request(
        request=request,
        msg_success=COMPENSATION_REMOVED_TEMPLATE.format(comp.identifier),
        redirect_url=reverse("intervention:detail", args=(id,)) + "#related_data",
    )


def report_view(request:HttpRequest, id: str):
    """ Renders the public report view

    Args:
        request (HttpRequest): The incoming request
        id (str): The id of the intervention

    Returns:

    """
    template = "intervention/report/report.html"
    intervention = get_object_or_404(Intervention, id=id)

    tab_title = _("Report {}").format(intervention.identifier)
    # If intervention is not recorded (yet or currently) we need to render another template without any data
    if not intervention.is_ready_for_publish():
        template = "report/unavailable.html"
        context = {
            TAB_TITLE_IDENTIFIER: tab_title,
        }
        context = BaseContext(request, context).context
        return render(request, template, context)

    # Prepare data for map viewer
    geom_form = SimpleGeomForm(
        instance=intervention
    )
    parcels = intervention.get_underlying_parcels()

    distinct_deductions = intervention.deductions.all().distinct(
        "account"
    )
    qrcode_url = request.build_absolute_uri(reverse("intervention:report", args=(id,)))
    qrcode_img = generate_qr_code(qrcode_url, 10)
    qrcode_lanis_url = intervention.get_LANIS_link()
    qrcode_img_lanis = generate_qr_code(qrcode_lanis_url, 7)

    context = {
        "obj": intervention,
        "deductions": distinct_deductions,
        "qrcode": {
            "img": qrcode_img,
            "url": qrcode_url,
        },
        "qrcode_lanis": {
            "img": qrcode_img_lanis,
            "url": qrcode_lanis_url,
        },
        "geom_form": geom_form,
        "parcels": parcels,
        TAB_TITLE_IDENTIFIER: tab_title,
    }
    context = BaseContext(request, context).context
    return render(request, template, context)
